@@ -2,7 +2,7 @@
  * This file is part of gspell, a spell-checking library.
  *
  * Copyright 2002 - Paolo Maggi
- * Copyright 2015, 2016 - Sébastien Wilmet
+ * Copyright 2015, 2016, 2017 - Sébastien Wilmet
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -22,7 +22,10 @@
  * Copyright 2002 - Evan Martin
  */
 
-#include "config.h"
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include "gspell-inline-checker-text-buffer.h"
 #include <string.h>
 #include <glib/gi18n-lib.h>
@@ -80,26 +83,13 @@ typedef enum
 /* Timeout durations in milliseconds. Writing and deleting text should be smooth
  * and responsive.
  */
-#define TIMEOUT_DURATION_BUFFER_MODIFIED 400
+#define TIMEOUT_DURATION_BUFFER_MODIFIED 16
 #define TIMEOUT_DURATION_DRAWING 20
 
+#define PERF_DEBUG FALSE
+
 G_DEFINE_TYPE (GspellInlineCheckerTextBuffer, _gspell_inline_checker_text_buffer, G_TYPE_OBJECT)
 
-/* FIXME properly. Workaround for bug in GtkTextView:
- * https://bugzilla.gnome.org/show_bug.cgi?id=763741
- */
-static void
-queue_draw_hack (GspellInlineCheckerTextBuffer *spell)
-{
-	GSList *l;
-
-	for (l = spell->views; l != NULL; l = l->next)
-	{
-		GtkWidget *view = GTK_WIDGET (l->data);
-		gtk_widget_queue_draw (view);
-	}
-}
-
 /* Remove the highlight_tag only if present. If gtk_text_buffer_remove_tag() is
  * called when the tag is not present, GtkTextView anyway queues a redraw, which
  * we want to avoid (it can lead to an infinite loop).
@@ -132,56 +122,9 @@ remove_highlight_tag_if_present (GspellInlineCheckerTextBuffer *spell,
 					    spell->highlight_tag,
 					    start,
 					    end);
-
-		queue_draw_hack (spell);
 	}
 }
 
-static void
-check_word (GspellInlineCheckerTextBuffer *spell,
-	    const GtkTextIter             *start,
-	    const GtkTextIter             *end)
-{
-	gchar *word;
-	GError *error = NULL;
-	gboolean correctly_spelled;
-
-	if (spell->spell_checker == NULL ||
-	    gspell_checker_get_language (spell->spell_checker) == NULL)
-	{
-		return;
-	}
-
-	if (!_gspell_text_iter_starts_word (start) ||
-	    !_gspell_text_iter_ends_word (end))
-	{
-		g_warning ("Spell checking: @start and @end must delimit a word");
-		return;
-	}
-
-	word = gtk_text_buffer_get_text (spell->buffer, start, end, FALSE);
-
-	correctly_spelled = gspell_checker_check_word (spell->spell_checker,
-						       word, -1,
-						       &error);
-
-	if (error != NULL)
-	{
-		g_warning ("Inline spell checker: %s", error->message);
-		g_clear_error (&error);
-	}
-
-	if (!correctly_spelled)
-	{
-		gtk_text_buffer_apply_tag (spell->buffer,
-					   spell->highlight_tag,
-					   start,
-					   end);
-	}
-
-	g_free (word);
-}
-
 static void
 adjust_iters (GtkTextIter *start,
 	      GtkTextIter *end,
@@ -194,14 +137,12 @@ adjust_iters (GtkTextIter *start,
 			    !_gspell_text_iter_starts_word (start))
 			{
 				_gspell_text_iter_backward_word_start (start);
-				g_assert (_gspell_text_iter_starts_word (start));
 			}
 
 			if (_gspell_text_iter_inside_word (end) &&
 			    !_gspell_text_iter_starts_word (end))
 			{
 				_gspell_text_iter_forward_word_end (end);
-				g_assert (_gspell_text_iter_ends_word (end));
 			}
 			break;
 
@@ -211,13 +152,11 @@ adjust_iters (GtkTextIter *start,
 			     !_gspell_text_iter_starts_word (start)))
 			{
 				_gspell_text_iter_backward_word_start (start);
-				g_assert (_gspell_text_iter_starts_word (start));
 			}
 
 			if (_gspell_text_iter_inside_word (end))
 			{
 				_gspell_text_iter_forward_word_end (end);
-				g_assert (_gspell_text_iter_ends_word (end));
 			}
 			break;
 
@@ -226,14 +165,73 @@ adjust_iters (GtkTextIter *start,
 	}
 }
 
+/* Free *attrs with g_free() when no longer needed. */
+static void
+get_pango_log_attrs (const gchar   *text,
+		     PangoLogAttr **attrs,
+		     gint          *n_attrs)
+{
+	*n_attrs = g_utf8_strlen (text, -1) + 1;
+	*attrs = g_new0 (PangoLogAttr, *n_attrs);
+
+	pango_get_log_attrs (text,
+			     strlen (text),
+			     -1,
+			     NULL,
+			     *attrs,
+			     *n_attrs);
+
+	_gspell_utils_improve_word_boundaries (text, *attrs, *n_attrs);
+}
+
+static gboolean
+should_apply_tag_to_misspelled_word (GspellInlineCheckerTextBuffer *spell,
+				     const GtkTextIter             *word_start,
+				     const GtkTextIter             *word_end)
+{
+	GtkTextIter iter;
+
+	if (spell->no_spell_check_tag == NULL)
+	{
+		return TRUE;
+	}
+
+	if (gtk_text_iter_has_tag (word_start, spell->no_spell_check_tag))
+	{
+		return FALSE;
+	}
+
+	iter = *word_start;
+	if (!gtk_text_iter_forward_to_tag_toggle (&iter, spell->no_spell_check_tag))
+	{
+		return TRUE;
+	}
+
+	return gtk_text_iter_compare (word_end, &iter) <= 0;
+}
+
+/* A first implementation of this function used the _gspell_text_iter_*()
+ * functions in a loop to navigate through words between @start and @end.
+ * But the _gspell_text_iter_*() functions are *slow*. So a new implementation
+ * has been written to reduce the number of calls to GtkTextView functions, and
+ * it's up to 20x faster! (200 ms -> 10 ms).
+ * And there is most probably still room for performance improvements.
+ */
 static void
 check_subregion (GspellInlineCheckerTextBuffer *spell,
 		 GtkTextIter                   *start,
 		 GtkTextIter                   *end)
 {
-	GtkTextIter word_start;
+	gchar *text;
+	const gchar *cur_text_pos;
+	const gchar *word_start;
+	gint word_start_char_pos;
+	PangoLogAttr *attrs;
+	gint n_attrs;
+	gint attr_num;
+	gint start_offset;
 
-	g_assert_cmpint (gtk_text_iter_compare (start, end), <=, 0);
+	g_return_if_fail (gtk_text_iter_compare (start, end) <= 0);
 
 	adjust_iters (start, end, ADJUST_MODE_STRICTLY_INSIDE_WORD);
 
@@ -242,68 +240,108 @@ check_subregion (GspellInlineCheckerTextBuffer *spell,
 				    start,
 				    end);
 
-	if (_gspell_text_iter_starts_word (start))
+	if (spell->spell_checker == NULL ||
+	    gspell_checker_get_language (spell->spell_checker) == NULL)
 	{
-		word_start = *start;
+		return;
 	}
-	else
+
+	text = gtk_text_iter_get_slice (start, end);
+
+	get_pango_log_attrs (text, &attrs, &n_attrs);
+
+	attr_num = 0;
+	cur_text_pos = text;
+	word_start = NULL;
+	word_start_char_pos = 0;
+
+	start_offset = gtk_text_iter_get_offset (start);
+
+	while (attr_num < n_attrs)
 	{
-		GtkTextIter next_word_end;
+		PangoLogAttr *cur_attr = &attrs[attr_num];
 
-		/* Based on adjust_iters() code. */
-		g_assert (!_gspell_text_iter_inside_word (start));
-
-		next_word_end = *start;
-		_gspell_text_iter_forward_word_end (&next_word_end);
-
-		/* Didn't move, there are no words after @start. */
-		if (gtk_text_iter_equal (&next_word_end, start))
+		if (word_start != NULL &&
+		    cur_attr->is_word_end)
 		{
-			return;
+			gint word_byte_length;
+			gboolean misspelled;
+			GError *error = NULL;
+
+			if (cur_text_pos != NULL)
+			{
+				word_byte_length = cur_text_pos - word_start;
+			}
+			else
+			{
+				word_byte_length = -1;
+			}
+
+			misspelled = !gspell_checker_check_word (spell->spell_checker,
+								 word_start,
+								 word_byte_length,
+								 &error);
+
+			if (error != NULL)
+			{
+				g_warning ("Inline spell checker: %s", error->message);
+				g_clear_error (&error);
+			}
+
+			if (misspelled)
+			{
+				gint word_start_offset;
+				gint word_end_offset;
+				GtkTextIter word_start_iter;
+				GtkTextIter word_end_iter;
+
+				word_start_offset = start_offset + word_start_char_pos;
+				word_end_offset = start_offset + attr_num;
+
+				gtk_text_buffer_get_iter_at_offset (spell->buffer,
+								    &word_start_iter,
+								    word_start_offset);
+
+				gtk_text_buffer_get_iter_at_offset (spell->buffer,
+								    &word_end_iter,
+								    word_end_offset);
+
+				if (should_apply_tag_to_misspelled_word (spell, &word_start_iter, &word_end_iter))
+				{
+					gtk_text_buffer_apply_tag (spell->buffer,
+								   spell->highlight_tag,
+								   &word_start_iter,
+								   &word_end_iter);
+				}
+			}
+
+			/* Find next word start. */
+			word_start = NULL;
 		}
 
-		g_assert (_gspell_text_iter_ends_word (&next_word_end));
-		g_assert_cmpint (gtk_text_iter_compare (start, &next_word_end), <, 0);
-
-		word_start = next_word_end;
-		_gspell_text_iter_backward_word_start (&word_start);
-		g_assert (_gspell_text_iter_starts_word (&word_start));
-
-		/* This assertion has failed in some cases, see:
-		 * https://bugzilla.gnome.org/show_bug.cgi?id=778883
-		 */
-		g_assert_cmpint (gtk_text_iter_compare (start, &word_start), <, 0);
-	}
-
-	while (_gspell_utils_skip_no_spell_check (spell->no_spell_check_tag, &word_start, end) &&
-	       gtk_text_iter_compare (&word_start, end) < 0)
-	{
-		GtkTextIter word_end;
-		GtkTextIter next_word_start;
-
-		g_assert (_gspell_text_iter_starts_word (&word_start));
-
-		word_end = word_start;
-		_gspell_text_iter_forward_word_end (&word_end);
-
-		g_assert_cmpint (gtk_text_iter_compare (&word_end, end), <=, 0);
-
-		check_word (spell, &word_start, &word_end);
-
-		next_word_start = word_end;
-		_gspell_text_iter_forward_word_end (&next_word_start);
-		_gspell_text_iter_backward_word_start (&next_word_start);
-
-		/* Make sure we've actually advanced (we don't advance if we
-		 * have just checked the last word of the buffer).
-		 */
-		if (gtk_text_iter_compare (&next_word_start, &word_start) <= 0)
+		if (word_start == NULL &&
+		    cur_attr->is_word_start)
 		{
+			word_start = cur_text_pos;
+			word_start_char_pos = attr_num;
+		}
+
+		if (cur_text_pos == NULL &&
+		    attr_num != n_attrs - 1)
+		{
+			g_warning ("%s(): problem in loop iteration, attr_num=%d but should be %d.",
+				   G_STRFUNC,
+				   attr_num,
+				   n_attrs - 1);
 			break;
 		}
 
-		word_start = next_word_start;
+		attr_num++;
+		cur_text_pos = g_utf8_find_next_char (cur_text_pos, NULL);
 	}
+
+	g_free (text);
+	g_free (attrs);
 }
 
 static void
@@ -378,7 +416,7 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 	}
 	else
 	{
-		g_assert_true (spell->unit_test_mode);
+		g_assert (spell->unit_test_mode);
 		gtk_text_buffer_get_bounds (spell->buffer, &visible_start, &visible_end);
 	}
 
@@ -432,6 +470,7 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 		GtkTextIter end;
 		GtkTextIter orig_start;
 		GtkTextIter orig_end;
+		gboolean bug = FALSE;
 
 		if (!_gspell_region_iter_get_subregion (&intersect_iter, &start, &end))
 		{
@@ -441,14 +480,52 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 		orig_start = start;
 		orig_end = end;
 
-		check_subregion (spell, &start, &end);
+		{
+#if PERF_DEBUG
+			GTimer *timer;
+
+			g_print ("check_subregion [%d, %d]\n",
+				 gtk_text_iter_get_offset (&start),
+				 gtk_text_iter_get_offset (&end));
+
+			timer = g_timer_new ();
+#endif
+
+			check_subregion (spell, &start, &end);
+
+#if PERF_DEBUG
+			g_print ("check_subregion took %lf ms.\n\n",
+				 1000 * g_timer_elapsed (timer, NULL));
+			g_timer_destroy (timer);
+#endif
+		}
 
 		/* Ensure that we don't have an infinite loop. We must subtract
-		 * from scan_region at least [start, end], otherwise we will
-		 * re-check the same subregion again and again.
+		 * from scan_region at least [orig_start, orig_end], otherwise
+		 * we will re-check the same subregion again and again.
 		 */
-		g_assert (gtk_text_iter_compare (&start, &orig_start) <= 0);
-		g_assert (gtk_text_iter_compare (&orig_end, &end) <= 0);
+		if (gtk_text_iter_compare (&orig_start, &start) < 0)
+		{
+			g_warning ("Should not reach this code path.");
+			bug = TRUE;
+			start = orig_start;
+		}
+		if (gtk_text_iter_compare (&end, &orig_end) < 0)
+		{
+			g_warning ("Should not reach this code path.");
+			bug = TRUE;
+			end = orig_end;
+		}
+
+		if (bug)
+		{
+			gchar *text;
+
+			text = gtk_text_iter_get_slice (&start, &end);
+			g_warning ("Text that caused the bug: '%s'", text);
+			g_warning ("Please report the bug to: " PACKAGE_BUGREPORT);
+			g_free (text);
+		}
 
 		_gspell_region_subtract_subregion (spell->scan_region, &start, &end);
 
@@ -461,8 +538,6 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 	{
 		g_clear_object (&spell->scan_region);
 	}
-
-	queue_draw_hack (spell);
 }
 
 static void
@@ -867,8 +942,6 @@ remove_tag_to_word (GspellInlineCheckerTextBuffer *spell,
 
 		iter = match_end;
 	}
-
-	queue_draw_hack (spell);
 }
 
 static void
@@ -1060,6 +1133,7 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 	GtkTextIter start;
 	GspellTextBuffer *gspell_buffer;
 	GspellChecker *checker;
+	GdkRGBA underline_color;
 
 	g_return_if_fail (GTK_IS_TEXT_BUFFER (buffer));
 	g_return_if_fail (spell->buffer == NULL);
@@ -1115,8 +1189,11 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 				 spell,
 				 G_CONNECT_AFTER);
 
+	_gspell_utils_init_underline_rgba (&underline_color);
+
 	spell->highlight_tag = gtk_text_buffer_create_tag (spell->buffer, NULL,
-							   "underline", PANGO_UNDERLINE_ERROR,
+							   "underline", PANGO_UNDERLINE_SINGLE,
+							   "underline-rgba", &underline_color,
 							   NULL);
 	g_object_ref (spell->highlight_tag);
 
